%% LessardEtAl_15SIAM_Figure5
% This file regenerates the rate bounds for Heavy Ball method as shown in
% Figure 5 of the 2015 SIAM paper by Lessard, Recht, and Packard.
%
% The condition ratio is k=L/m.  All calculations are performed
% assuming m=1 so that k=L.  This assumption is without loss of
% generality (for strongly convex functions) by a simple re-scaling.
% Specifically, consider the optimization min_x f(x) where
% f in S(m,L) with m>0 (strongly convex).  Rescale the objective
% function to fhat := f/m so that fhat in S(1,k). The scaled optimization
% min_x fhat(x) has the same minima.

%% Flag to use CVX or LMILab
% LMILab is significantly faster on these problems but the code is 
% not as readable. Also, the LMILab code has been written for the general
% rho-Hard Zames-Falb condition.
useCVX = false;

%% Theoretical Lower Bound
% See Eq 4.9 in Section 4.5
N1 = 200;
k1 = logspace(0,3,N1);
rhoTheoryLB = ( sqrt(k1)-1 ) ./ ( sqrt(k1)+1 );

%% Optimal Gradient Rate
% The optimal tuning is
%  alpha = 2/(L+m)
% The bound with arbitrary functions using the optimal tuning is given in
% Prop. 12 in Section 4.3.  They cite work by Nesterov for this bound.
rhoOptimalGrad = (k1-1)./(k1+1);

%% Sector IQC
N2 = 10;
k2 = logspace(0,log10(50),N2);
rhoSector = zeros(N2,1);
for i1 = 1:N2
    % Select condition # and set algorithm parameters
    % For Heavy ball alpha = 4/( sqrt(L) + sqrt(m) )^2. In the
    % code below assume m=1 so that L=k.
    ki = k2(i1);
    L = ki;
    beta = ( (sqrt(ki) -1 ) / (sqrt(ki)+1) )^2;
    alpha = 4/( sqrt(ki) +1 )^2;
    
    % Plant data for Heavy Ball
    % (This form employs dimensionality reduction)
    A = [1+beta -beta; 1 0];
    B = [-alpha;0];
    C = [1 0];
    D = 0;
    G = ss(A,B,C,D,1);
    
    % Bisection
    if useCVX
        rhoSector(i1) = rateboundSector( G, ki);
    else
        Nzf = 0; L=ki; m=1;
        rhoSector(i1) = ratebound( G, L, m, Nzf);
    end
end


%% Weighted Off-by-1 IQC
N3 = 50;
k3 = logspace(0,4,N3);
rhoOffBy1 = zeros(N3,1);
Popt = zeros(3,3,N3);
Copt = zeros(3,3,N3);
hopt = zeros(N3,1);
LMIopt = zeros(4,4,N3);
Pare = Popt;
for i1 = 1:N3
    % Select condition # and set algorithm parameters
    % For Heavy ball alpha = 4/( sqrt(L) + sqrt(m) )^2. In the
    % code below assume m=1 so that L=k.
    ki = k3(i1);
    L = ki;
    beta = ( (sqrt(ki) -1 ) / (sqrt(ki)+1) )^2;
    alpha = 4/( sqrt(ki) +1 )^2;
    
    % Plant data for Heavy Ball
    % (This form employs dimensionality reduction)
    A = [1+beta -beta; 1 0];
    B = [-alpha;0];
    C = [1 0];
    D = 0;
    G = ss(A,B,C,D,1);
    
    % Bisection
    if useCVX
        rhoOffBy1(i1) = rateboundOffBy1( G, ki);
    else
        Nzf = 1; L=ki; m=1;
        rhoOffBy1(i1) = ratebound( G, L, m, Nzf);
    end
end

%% Plot Results
figure(1);
semilogx(k2,rhoSector,'k-s',k3,rhoOffBy1,'k-o',...
    k1,rhoOptimalGrad,'r--',k1,rhoTheoryLB,'b--');
xlabel('Condition ratio L/m');
ylabel('Convergence rate \rho');
axis([1 1e3 0 1.1]);

legend('LMI (sector)','LMI (wted off-by-1)',...
    'Optimal Gradient','Theoretical lower bound','Location','Southeast');
grid on;


%% Plot Results: Without Gradient Curve
figure(2);
ph = semilogx(k2,rhoSector,'r:s',k3,rhoOffBy1,'k-o',k1,rhoTheoryLB,'b--');
xlabel('Condition ratio L/m');
ylabel('Convergence rate \rho');
axis([1 1e3 0 1.1]);

legend('LMI: Sector','LMI: Weighted Off-by-1',...
    'Nesterov Lower Bound','Location','Southeast');
grid on;
set(ph,'LineWidth',2)
set(gca,'FontSize',12);


%% Plot Results: Sector And Nesterov lower bound
figure(3);
ph = semilogx(k2,rhoSector,'r:s',k1,rhoTheoryLB,'b--');
xlabel('Condition ratio L/m');
ylabel('Convergence rate \rho');
axis([1 1e3 0 1.1]);

legend('LMI: Sector','Nesterov Lower Bound','Location','Southeast');
grid on;
set(ph,'LineWidth',2)
set(gca,'FontSize',12);

