function [rho,P,h] = ratebound(G,L,m,Nzf,opt)
% [rho,P,lam] = ratebound(G,L,m,Nzf)
%
% Uses the *causal* Zames-Falb multiplier to computes convergence rate
% bound rho for lft(phi,G) where:
%   1) G is a discrete-time, SISO, LTI system.
%   2) phi is the gradient of a function f in S(m,L), i.e. f is strongly
%      convex with parameter m and has gradient with Lipschitz bound L.
%   3) Nzf>0 is the order of the causal Zames-Falb. Nzf=0 is the sector
%      IQC and Nzf=1 is the weighted off-by-1 IQC. In general, Nzf>0
%      contains all off-by-j terms up through j=Nzf.
%
% See the following reference for details:
%   Lessard, Recht, Packard, Analysis and Design of Optimization
%   Algorithms via Integral Quadratic Constraints, SIAM, 2015.

%% Bisection Options (Optional Input)
narginchk(4,5)
if nargin == 5
    rholb = opt(1);
    rhoub = opt(2);
    rhotol = opt(3);
else
    % Default Values
    rholb = 0;
    rhoub = 2;
    rhotol = 1e-3;
end

% Flag to display debugging information
dispflag = false;

%% IQC Filter and Center Matrix
S = [L -1; -m 1];
if Nzf ==0
    Apsi = [];
    Bpsi = zeros(0,2);
else
    Apsi = diag(ones(Nzf-1,1),-1);
    Bpsi = [-1 0; zeros(Nzf-1,2)]*S;
end
Dpsi = S;
M = [0 1; 1 0];

%% Extended system
[Ag,Bg,Cg,Dg] = ssdata(G);
Ng = size(Ag,1);

Ae = [Ag zeros(Ng,Nzf); Bpsi(:,1)*Cg Apsi];
Be = [Bg; Bpsi(:,2)+Bpsi(:,1)*Dg];
De = Dpsi(:,1)*Dg+Dpsi(:,2);
Ne = Ng+Nzf;

% For Zames-Falb (with Nzf>0) the output matrix is
%   Cpsi = [h1 ... hNzf; 0 ... 0]
% where hi are decision variables subject to the constraints
%   h1/rho^2 + ... + hNzf/ rho^(2Nzf) <=1
%   h1>=0, ..., hNzf>=0
% Split the extended output matrix as:
%   Ce = [Dpsi1*C Cpsi] = Ce0 + Ceh
%   where Ceh := [ 0_{1-by-Ng} h; 0_{1-by-(Ng+Nzf)}]
Ce0 = [Dpsi(:,1)*Cg zeros(2,Nzf)];

%% Initialize LMI

% Set Empty LMI
setlmis([]);

% Create Variables
% P: Ne-by-Ne storage matrix
% h: 1-by-Nzf vector of coefficients weighting each off-by-j term.
% CDeh = [Ceh 0*De]
P = lmivar(1,[Ne, 1]);
if Nzf>0
    [h,~,hdec] = lmivar(2,[1 Nzf]);
    CDeh = lmivar(3,[zeros(1,Ng) hdec 0; zeros(1,Ne) 0]);
end

% Enforce P>0
lmiterm([-1 1 1 P],1,1);

% Store (partial) LMI information
lmisys0 = getlmis;

%% Perform Bisection
Popt = []; hopt = [];
while rhoub-rholb>rhotol
    % Select rho
    rho = (rholb+rhoub)/2;
    
    % Main LMI
    % (.)' P [Ae Be] - rho^2 blkdiag(P,0) + (.)' M [Ce De] <=0
    % Note that CeB(h)' M CeB(h) = 0.
    setlmis(lmisys0)
    lmiterm([2 1 1 P],[Ae Be]',[Ae Be]);
    Rfac = [rho*eye(Ne) zeros(Ne,1)];
    lmiterm([-2 1 1 P],Rfac',Rfac);
    lmiterm([2 1 1  0],[Ce0 De]'*M*[Ce0 De]);
    if Nzf>0
        lmiterm([2 1 1  CDeh],[Ce0 De]'*M,1,'s');
    end
    
    if Nzf>0
        % Constraint: h1/rho^2 + ... + hNzf/ rho^(2Nzf) <=1
        % Implemented as h*vrho + vrho'*h' <=2
        % where vrho = [1/rho^2; ...; 1/rho^(2Nzf)]
        vrho = (rho^-2).^(1:Nzf);
        lmiterm([-3 1 1 0],2);
        lmiterm([3 1 1 h],1,vrho(:),'s');
        
        % Constraint: hi>=0
        % Implemented as h*ei + ei'*h' >=0 where ei is i^th basis vec
        for i=1:Nzf
            ei = zeros(Nzf,1);
            ei(i) = 1;
            lmiterm([-(3+i) 1 1 h],1,ei,'s');
        end
    end
    
    % Solve Feasibility problem
    ttol = -1e-10;
    opt = [0 0 0 0 1];
    lmisys = getlmis;
    [tmin,xfeas] = feasp(lmisys,opt,10*ttol);
    
    solveflag = ( tmin<= ttol );
    if solveflag
        rhoub = rho;
        Popt = dec2mat(lmisys,xfeas,P);
        if Nzf>0
            hopt = dec2mat(lmisys,xfeas,h);
        end
    else
        rholb = rho;
    end
    if dispflag
        fprintf([' rho = %4.3f \t Newlb = %4.3f \t Newub = ' ...
            '%4.3f \t tmin = %.3e \t Solve = %d \n'],...
            rho,rholb,rhoub,tmin,solveflag);
    end
end
if dispflag
    fprintf('\n');
end

%% Store results
rho = rhoub;
P = Popt;
if Nzf
    h = hopt;
else
    h = [];
end

%% Check LMIs (For Debugging)
% Cpsi = [h; zeros(1,Nzf)];
% Ce = [Dpsi(:,1)*Cg Cpsi];
% LMI = [Ae Be]'*P*[Ae Be] - rho^2*blkdiag(P,0) + [Ce De]'*M*[Ce De];
% keyboard
